

//
// linked lists
//


// linear linked list functions
namespace llist
{
//
// generic linked list functions
//
 template <class CLSA,class CLSB >
 inline void add_head(CLSA * &l, CLSB * ob)
 {
  ob->next = l;
  l     = ob;
 }
#if 0
 template <class T,class TD >
 inline void add_list_tail(T * &l, TD * ob)
 {
   T ** lp = &l;
  // skip to end
  for(;*lp;lp = &(*lp)->next );
  *lp = ob;
 }
 template <class T,class TD >
 inline void add_tail(T * &l, TD * ob)
 {
   ob->next =0;
   add_list_tail(l,ob);
 }
#endif

 template <class T >
 inline T * find_last(T * l)
 {
   if( !l ) return 0;
   for(;l->next;l = l->next) ;
   return l;
 }


 template <class T,class TD >
 inline T ** find(T * &l, TD * a)
 {
   T ** rv = &l;
   for(;*rv; rv = &(*rv)->next )
   {
	   if(*rv == a)
	    return rv;
   }
   return 0;
 }

 template <class T>
 inline T ** find_lastp(T * &l )
 {
   T ** rv = &l;
   for(;*rv; rv = &(*rv)->next );
   return rv;
 }
 template <class T,class TD >
 inline void add_list_tail(T * &l, TD * ob)
 {
  *find_lastp(l) = ob;
 }
 template <class T,class TD >
 inline void add_tail(T * &l, TD * ob)
 {
   ob->next =0;
   add_list_tail(l,ob);
 }


 // uses operator == 
 template <class T,class TD >
 inline T ** find_eq(T * &l, const TD & a)
 {
   T ** rv = &l;
   for(;*rv; rv = &(*rv)->next )
   {
	   if(**rv == a)
	    return rv;
   }
   return 0;
 }

 template <class T,class TD >
 inline void remove(T * &l,TD * a)
 {
   T ** lp = find(l,a);
   assert(lp);
   *lp = a->next;
 }
 template <class T,class TD >
 inline void remove_if_present(T * &l,TD * a)
 {
   T ** lp = find(l,a);
   if( lp )
    *lp = a->next;
 }
 template <class T,class TD >
 inline void add_head_if_absent(T * &l,TD * a)
 {
  if( find(l,a) == 0 )
    add_head(l,a);
 }



 template <class  T>
 inline int size(const T * l)
 {
  int rv = 0;
  for(const T * tmp = l;tmp;tmp = tmp->next  )
    ++rv;
  return rv;
 }

 template <class  T>
 inline T * index(T * lst, int idx )
 {
  while(--idx >= 0)
  {
   assert( lst );
   lst = lst->next;
  }

  assert( lst );
  return lst;
 }

 template <class  T>
 inline int index(const T * lst,const T * item )
 {
  int rv =0;
  while( lst != item )
  {
   ++rv;
   assert( lst );
   lst = lst->next;
  }
  return rv;
 }


 template <class T> 
 struct tail_inserter
 {
  T ** insp;
  inline tail_inserter( T * & list ) { insp = &list; }
  inline void add( T * item ) { *insp = item; insp = &item->next; }
  inline void end() { *insp = 0; }
 };
 
 // forward iterator
 template <class T> 
 struct itf
 {
  T * cur;
  
  inline itf(T * list) : cur(list) {}
  
  inline T * operator ->() { return cur;}
  inline T * p()           { return cur;}
  inline void operator ++ () { assert(cur); cur = cur->next; }
  inline int operator !() const  { return cur != 0; }
  static inline T * type_tag() { return 0; }
 };

 template <class T> 
 struct itf_cnt : itf<T>
 {
  int cnt; //  keeps track of count
  
  inline itf_cnt(T * list) : itf<T>(list), cnt(0) {}
  inline void operator ++ () { ++cnt; itf<T>::operator ++(); }
 };


}


// doubly linked list functions
namespace dlist
{
//
// generic linked list functions
//
 template <class  TA,class  TB >
 inline void link(TA * a, TB * b)
 {
  a->next = b;
  b->prev = a;
 }

 template <class T >
 inline void remove(T * a)
 {
  link(a->prev,a->next);
 }
 template <class TA,class TB >
 inline void insert_prev(TA * ins, TB * already_in) 
 {
  link(already_in->prev,ins);
  link(ins,already_in);
 }
 template <class TA,class TB >
 inline void insert_next(TA * ins, TB * already_in) 
 {
  link(ins,already_in->next);
  link(already_in,ins);
 }


}











//
// forward linked list (can only move in 1 direction)
//
template <class T>
struct fllist
{
 T * l;
 
 inline fllist(T * _l = 0): l(_l) {}

 inline operator T * () const { return l; }

 inline void transfer_from( fllist<T> & b);

 inline void add_head( T * ob );
 inline void add_tail( T * ob );
 inline void add_head_if_absent( T * ob );
 inline void remove_if_present( T * ob );
 
 inline void remove( T * ob );
 inline T ** find( T * ob );
};


#define DECF(RV)  template<class T> inline RV fllist<T>::


DECF(void) add_head( T * ob ) {  llist::add_head(l,ob);  }
DECF(void) add_tail( T * ob ) {  llist::add_tail(l,ob);  }
DECF(void) add_head_if_absent( T * ob ) {  llist::add_head_if_absent(l,ob);  }
DECF(void) remove_if_present( T * ob )  {  llist::remove_if_present(l,ob);  }
DECF(void) remove( T * ob )   {  llist::remove(l,ob);  }
DECF(T**)  find( T * ob )     {  return llist::find(l,ob);  }
DECF(void) transfer_from( fllist<T> & b) 
 {
  assert( (l==0)&&(b.l) );
  l = b.l; b.l =0;
 }



#undef DECF



